import datetime
import shlex
import textwrap

import awscli
import awscli.clidriver as clidriver
import botocore
import botocore.hooks
from awscli.plugin import load_plugins

from .utils import convert_from_camelcase, json_dumps


class Trace:
    def __init__(self):
        self.request_id = "<not set>"
        self.fn_name = "<not set>"
        self.inparams = {}
        self.outparams = {}
        self.ts_start = None
        self.ts_end = None

    def start(self):
        self.ts_start = datetime.datetime.now()

    def finish(self):
        self.ts_end = datetime.datetime.now()

    def get_output_value(self, name):
        names = name.split(".")
        d = self.outparams
        for name in names:
            if name not in d:
                return None
            d = d[name]
        return d

    def set_input(self, fn_name, params):
        self.fn_name = fn_name
        self.inparams = params

    def set_output(self, request_id, fn_name, values):
        if fn_name != self.fn_name:
            raise ValueError("unexpected call with non-matching function names")
        self.fn_name = fn_name
        self.request_id = request_id
        self.outparams = values

    def to_dict(self):
        return {
            "request_id": self.request_id,
            "fn_name": self.fn_name,
            "inparams": self.inparams,
            "outparams": self.outparams,
            "ts_start": self.ts_start,
            "ts_end": self.ts_end
        }

    @staticmethod
    def from_dict(d):
        obj = Trace()
        for n in ("request_id", "fn_name", "inparams", "outparams", "ts_start", "ts_end"):
            if n not in d:
                raise ValueError("invalid input")
            setattr(obj, n, d[n])
        return obj

    def get_shell_var(self, name, val):
        optname = shlex.quote("--{}".format(convert_from_camelcase(name)))

        if type(val) == dict or type(val) == list:
            val = json_dumps(val, pretty=False)
        elif type(val) == str:
            val = str(val)
        else:
            raise NotImplementedError("{}".format(type(val)))

        optval = shlex.quote(val)
        return (optname, optval)

    def get_shell_poc(self, replace_vars={}):
        parts = self.fn_name.split(".")
        if len(parts) != 2:
            raise ValueError("invalid fn_name")
        resource, fn = parts

        args = ["aws", resource, convert_from_camelcase(fn)]
        for n in self.inparams:
            var = self.inparams[n] if n not in replace_vars else replace_vars[n]
            optname, optval = self.get_shell_var(n, var)
            args.append(optname)
            args.append(optval)
        return " ".join(args)

    def __str__(self):
        in_str, out_str = [], []
        for i in self.inparams:
            in_str.append("{}: {}".format(i, self.inparams[i]))
        for i in self.outparams:
            out_str.append("{}: {}".format(i, self.outparams[i]))

        return "Trace\n  ID: {} \n  API: {}\n  IN:\n{}\n  OUT:\n{}\n".format(
            self.request_id, self.fn_name,
            textwrap.indent("\n".join(in_str), "    "),
            textwrap.indent("\n".join(out_str), "    "))


class EventCapturer(botocore.hooks.HierarchicalEmitter):
    def __init__(self):
        super().__init__()
        self.trace = Trace()
        self.events_captured = []

    def emit(self, event_name, **kwargs):
        if event_name.startswith("provide-client-params"):
            fn_name = event_name[len("provide-client-params") + 1:]
            if len(fn_name) == 0:
                raise ValueError("unexpected fn name")
            if "params" not in kwargs:
                raise ValueError("unexpected input")
            params = kwargs["params"]
            self.trace.start()
            self.trace.set_input(fn_name, params)
        elif event_name.startswith("after-call"):
            fn_name = event_name[len("after-call") + 1:]
            if len(fn_name) == 0:
                raise ValueError("unexpected fn name")
            if "parsed" not in kwargs:
                raise ValueError("unexpected input")
            parsed = kwargs["parsed"]
            if "ResponseMetadata" not in parsed:
                raise ValueError("unexpected input")
            if "RequestId" not in parsed["ResponseMetadata"]:
                raise ValueError("unexpected input")
            req_id = parsed["ResponseMetadata"]["RequestId"]
            del parsed["ResponseMetadata"]
            self.trace.set_output(req_id, fn_name, parsed)
            self.trace.finish()
        self.events_captured.append(event_name)
        return super().emit(event_name, **kwargs)


class TraceRunner:
    def __init__(self):
        pass

    def run_aws_cmd(self, args):
        try:
            ev = EventCapturer()
            driver = self._create_clidriver(ev)
            retval = driver.main(args=args)
            # command failed so we don't record this trace and
            # can now simply bail out
            if retval != 0:
                return None
        except Exception as e:
            print("unknown exception occured: {}".format(str(e)))
            return None
        return ev.trace

    def _create_clidriver(self, ev):
        session = botocore.session.Session(awscli.EnvironmentVariables)
        # registering the event emitter needs to be done here immediately to
        # prevent argparsing errors
        session.register_component("event_emitter", ev)
        awscli.clidriver._set_user_agent_for_session(session)
        load_plugins(session.full_config.get('plugins', {}),
                     event_hooks=session.get_component('event_emitter'))
        driver = clidriver.CLIDriver(session=session)
        return driver



import io
import tempfile
import unittest
import uuid
from datetime import datetime


class TestTracerUtils(unittest.TestCase):

    def test_datetime_json(self):
        try:
            from awstracer.utils import json_loads, json_load, json_dumps
        except Exception:
            self.fail("json_load(s), json_dump(s) import failed")
        val = {"_": datetime.now()}
        ret1 = json_dumps(val)
        self.assertNotEqual(ret1.find("_isoformat"), -1)
        ret2 = json_loads(ret1.encode("utf-8"))
        self.assertEqual(val, ret2)
        self.assertEqual(val, json_load(io.StringIO(ret1)))

    def test_camelcase_conversion(self):
        try:
            from awstracer.utils import convert_from_camelcase, convert_to_camelcase
        except Exception:
            self.fail("camelcase imports failed")

        test = "ExecutePartiQLStatement"
        ret1 = convert_from_camelcase(test)
        self.assertEqual("execute-parti-ql-statement", ret1)
        ret2 = convert_to_camelcase(ret1)

        # can't reliably convert this back
        self.assertNotEqual(test, ret2)

        test = "BlaBlaBla"
        ret1 = convert_from_camelcase(test)
        self.assertEqual("bla-bla-bla", ret1)
        ret2 = convert_to_camelcase(ret1)
        self.assertEqual(test, ret2)

    def test_process_file_argument(self):
        try:
            from awstracer.utils import process_file_argument
        except Exception:
            self.fail("process_file_argument failed")

        self.assertEqual(process_file_argument("bla"), "bla")
        # check non-existing file
        self.assertIsNone(process_file_argument("file://{}".format(uuid.uuid4().hex)))
        # test actual reading from file
        tp = tempfile.NamedTemporaryFile()
        tp.write(b"hello w0rld")
        tp.seek(0)
        arg = process_file_argument("file://{}".format(tp.name))
        self.assertEqual(arg, "hello w0rld")



import argparse
import os
import readline
import shlex

from .tracer import TraceRunner
from .utils import json_dumps, process_file_argument


class TraceRecorder(TraceRunner):
    def __init__(self, filename, prompt_on_save=True):
        super().__init__()
        if filename is None:
            raise ValueError("need a filename to save to")
        self._filename = filename
        self.traces = []
        self.prompt_on_save = prompt_on_save

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.prompt_on_save:
            if not confirm_prompt("Save cached trace to {}?".format(self._filename)):
                return
        with open(self._filename, "wb") as fd:
            data = []
            for trace in self.traces:
                data.append(trace.to_dict())
            data = json_dumps(data, pretty=True)
            fd.write(data.encode("utf-8"))

    def process_file_arguments(self, args):
        ret = []
        for arg in args:
            arg_ret = process_file_argument(arg)
            if not arg_ret:
                print("Couldn't read {}".format(arg))
                return None
            ret.append(arg_ret)
        return ret

    def run_aws_cmd(self, args):
        args = self.process_file_arguments(args)
        if not args:
            return
        trace = super().run_aws_cmd(args)
        if not trace:
            return
        save = True
        if self.prompt_on_save:
            save = confirm_prompt("Add command to trace cache?")
        if save:
            self.traces.append(trace)


def confirm_prompt(prompt):
    while True:
        answer = input("{} [y/N]: ".format(prompt)).lower()
        if answer == "" or answer == "n":
            return False
        elif answer == "y":
            return True
        else:
            continue


def run_misc_cmd(cmd, prompt=True):
    if prompt:
        if not confirm_prompt("Arbitrary shell command. Would you like to execute this anyway?"):
            return
    os.system(cmd)


def run(recorder, prompt_color=True, prompt_on_misc=True, enable_misc_cmd=True):
    try:
        readline.read_init_file()
    except Exception:
        # for some reason this fails on MacOS
        pass
    prompt = "\x1b[31m(rec)\x1b[0m " if prompt_color else "(rec) "
    multiline_prompt = "\x1b[31m>\x1b[0m " if prompt_color else "> "
    try:
        while True:
            cmd = input(prompt)

            # parse multiline commands and display a different prompt to denote
            # we're in multiline mode similarly to how bash does it
            if len(cmd) > 0 and cmd[-1] == "\\":
                parts = [cmd[:-1]]
                while True:
                    part = input(multiline_prompt)
                    if len(part) > 0 and part[-1] == "\\":
                        parts.append(part[:-1])
                        continue
                    parts.append(part)
                    break
                cmd = " ".join(parts)

            split_cmd = shlex.split(cmd)
            if len(split_cmd) == 0:
                continue
            if split_cmd[0] == "aws":
                recorder.run_aws_cmd(split_cmd[1:])
            else:
                if not enable_misc_cmd:
                    print("Can only record awscli commands.")
                    continue
                run_misc_cmd(cmd, prompt_on_misc)
    except EOFError:
        pass
    except KeyboardInterrupt:
        pass
    print("")


def opt_parser(args=None):
    parser = argparse.ArgumentParser(description="AWS CLI Trace Recorder")
    parser.add_argument("-c", action="store_false", dest="prompt_color", help="Turn off colorized output")
    parser.add_argument("-d", action="store_false", dest="prompt_on_misc", help="Do not ask for confirmation for shell execute")
    parser.add_argument("-n", action="store_false", dest="prompt_on_save", help="Do not prompt when adding to/saving trace files")
    parser.add_argument("-s", action="store_true", dest="enable_misc_cmd", help="Enable execution of all shell commands")
    group = parser.add_argument_group("required arguments")
    group.add_argument("--trace-file", metavar="FILE", type=str, required=True, help="output trace file", dest="trace_file")
    ns = parser.parse_args() if not args else parser.parse_args(args)
    return ns


def main():
    ns = opt_parser()

    with TraceRecorder(ns.trace_file, ns.prompt_on_save) as recorder:
        run(recorder,
            prompt_color=ns.prompt_color,
            prompt_on_misc=ns.prompt_on_misc,
            enable_misc_cmd=ns.enable_misc_cmd)


if __name__ == "__main__":
    main()



import io
import unittest
from contextlib import redirect_stderr, redirect_stdout


class TestPlayer(unittest.TestCase):
    def test_options(self):
        try:
            from awstracer.player import opt_parser
        except Exception:
            self.fail("cannot import opt_parser")
        with self.assertRaises(SystemExit):
            with redirect_stdout(io.StringIO()):
                with redirect_stderr(io.StringIO()):
                    opt_parser([])
                    # sleep delay cannot be negative so should exit
                    opt_parser(["--trace-file", "bla", "-s", "-1"])
        ns = opt_parser(["--trace-file", "bla"])
        self.assertEqual(ns.trace_file, "bla")
        # check default settings for options
        self.assertIsNone(ns.endpoint)
        self.assertIsNone(ns.params)
        self.assertIsNone(ns.profile)
        self.assertIsNone(ns.region)
        self.assertFalse(ns.dryrun)
        self.assertFalse(ns.debug)
        self.assertIsNone(ns.sleep_delay)
        self.assertTrue(ns.colorize)
        self.assertTrue(ns.stop_on_error)
        ns = opt_parser(["--trace-file", "bla", "--dryrun", "--region", "bl1", "--profile", "bl2", "--endpoint", "bl3", "-s", "2", "-d", "-f", "-c"])
        self.assertTrue(ns.dryrun)
        self.assertEqual(ns.region, "bl1")
        self.assertEqual(ns.profile, "bl2")
        self.assertEqual(ns.endpoint, "bl3")
        self.assertEqual(ns.sleep_delay, 2)
        self.assertTrue(ns.debug)
        self.assertFalse(ns.colorize)
        self.assertFalse(ns.stop_on_error)
        ns = opt_parser(["--trace-file", "bla", "--param", "p1", "val1", "--param", "p2", "val2"])
        self.assertIsInstance(ns.params, list)
        self.assertEqual(len(ns.params), 2)
        self.assertEqual(ns.params[0][0], "p1")
        self.assertEqual(ns.params[0][1], "val1")
        self.assertEqual(ns.params[1][0], "p2")
        self.assertEqual(ns.params[1][1], "val2")
        ns = opt_parser(["--trace-file", "bla", "-p", "p1", "val1", "-p", "p2", "val2"])
        self.assertIsInstance(ns.params, list)
        self.assertEqual(len(ns.params), 2)
        self.assertEqual(ns.params[0][0], "p1")
        self.assertEqual(ns.params[0][1], "val1")
        self.assertEqual(ns.params[1][0], "p2")
        self.assertEqual(ns.params[1][1], "val2")

    def test_player_class(self):
        try:
            from awstracer.player import TracePlayer
            from awstracer.tracer import Trace
            from awstracer.utils import json_dumps
        except Exception:
            self.fail("cannot import TracePlayer")
        with self.assertRaises(TypeError):
            with TracePlayer():
                pass
        inp = io.StringIO()
        inp.write("[]")
        inp.seek(0)
        with TracePlayer(inp, {}) as tp:
            self.assertEqual(len(tp.traces), 1)

        # check missing arguments
        inp = io.StringIO()
        inp.write("[{}]")
        inp.seek(0)
        with self.assertRaises(ValueError):
            with TracePlayer(inp, {}):
                pass

        t = Trace()
        t.start()
        t.set_input("bla.wut", [])
        t.set_output("reqid", "bla.wut", [])
        t.finish()
        inp = io.StringIO()
        inp.write("{}".format(json_dumps([t.to_dict()])))
        inp.seek(0)
        with TracePlayer(inp) as tp:
            self.assertEqual(len(tp.traces), 2)

        inp.seek(0)
        with TracePlayer(inp, profile="bla-profile", endpoint="bla-endpoint", region="bla-region") as tp:
            self.assertEqual(len(tp.traces), 2)
            out = io.StringIO()
            with redirect_stdout(out):
                tp.play_trace(dryrun=True, sleep_delay=0)
            val = out.getvalue()
            self.assertNotEqual(val.find("--profile bla-profile"), -1)
            self.assertNotEqual(val.find("--endpoint bla-endpoint"), -1)
            self.assertNotEqual(val.find("--region bla-region"), -1)

    def test_player_methods_existence(self):
        try:
            from awstracer.player import TracePlayer
        except Exception:
            self.fail("cannot import TracePlayer")
        methods = ["play_trace", "play_single_trace", "get_shell_poc", "find_connections_between_traces", "find_connections", "prune_connections"]
        for m in methods:
            self.assertIn(m, dir(TracePlayer))

    def test_player_edges(self):
        try:
            from awstracer.player import Edge, MatchingNameAndValueEdge, MatchingNameEdge, MatchingValueEdge
        except Exception:
            self.fail("cannot import TracePlayer")
        self.assertIsInstance(MatchingNameAndValueEdge.__bases__[0], type(Edge))
        self.assertIsInstance(MatchingNameEdge.__bases__[0], type(Edge))
        self.assertIsInstance(MatchingValueEdge.__bases__[0], type(Edge))

    def test_player_play_trace(self):
        try:
            from awstracer.player import TracePlayer
            from awstracer.tracer import Trace
            from awstracer.utils import json_dumps
        except Exception:
            self.fail("cannot import TracePlayer")

        f = io.StringIO()
        with redirect_stdout(f):
            t = Trace()
            t.start()
            t.set_input("bla.wut", {"arg1": "val'\"`ue1"})
            t.set_output("reqid", "bla.wut", [])
            t.finish()
            inp = io.StringIO()
            inp.write("{}".format(json_dumps([t.to_dict()])))
            inp.seek(0)
            with TracePlayer(inp, {}, prompt_color=False) as tp:
                tp.play_trace(dryrun=True, stop_on_error=True, sleep_delay=0)
        self.assertNotEqual(f.getvalue().find("(play) aws bla wut"), -1)
        self.assertNotEqual(f.getvalue().find("--arg1 'val'\"'\"'\"`ue1'"), -1)

    def test_player_find_and_prune_connections(self):
        try:
            from awstracer.player import TracePlayer
            from awstracer.tracer import Trace
            from awstracer.utils import json_dumps
            from awstracer.player import MatchingNameAndValueEdge, MatchingNameEdge, MatchingValueEdge
        except Exception:
            self.fail("cannot import TracePlayer")

        f = io.StringIO()
        with redirect_stdout(f):
            t = Trace()
            t.start()
            t.set_input("bla.wut", {"arg1": "val'\"`ue1"})
            t.set_output("reqid2", "bla.wut", {"ret1": "val1"})
            t.finish()
            t2 = Trace()
            t2.start()
            t2.set_input("bla.wut2", {"ret1": "val1"})
            t2.set_output("reqid2", "bla.wut2", {})
            t2.finish()
            t3 = Trace()
            t3.start()
            t3.set_input("bla.wut3", {"other-param-same-val": "val1"})
            t3.set_output("reqid3", "bla.wut3", {"output-value": "val1"})
            t3.finish()
            t4 = Trace()
            t4.start()
            t4.set_input("bla.wut4", {"output-value": "val2"})
            t4.set_output("reqid4", "bla.wut4", {})
            t4.finish()
            tt = [t, t2, t3, t4]
            inp = io.StringIO()
            inp.write("{}".format(json_dumps([x.to_dict() for x in tt])))
            inp.seek(0)
            with TracePlayer(inp, {}, prompt_color=False) as tp:
                tp.find_connections()
                self.assertEqual(len(tp.connections), 3)

                # test the first connection matching name+value
                c = tp.connections[0]
                self.assertIsInstance(c, MatchingNameAndValueEdge)
                self.assertEqual(c.trace_from.fn_name, "bla.wut")
                self.assertEqual(c.trace_to.fn_name, "bla.wut2")
                self.assertIn("ret1", c.trace_from.outparams)
                self.assertEqual(c.trace_from.outparams["ret1"], "val1")
                self.assertIn("ret1", c.trace_to.inparams)
                self.assertEqual(c.trace_to.inparams["ret1"], "val1")

                # test second connection which should be matching value only
                c = tp.connections[1]
                self.assertIsInstance(c, MatchingValueEdge)
                self.assertEqual(c.trace_from.fn_name, "bla.wut")
                self.assertEqual(c.trace_to.fn_name, "bla.wut3")
                self.assertIn("other-param-same-val", c.trace_to.inparams)
                self.assertEqual(c.trace_to.inparams["other-param-same-val"], "val1")
                self.assertIn("ret1", c.trace_from.outparams)
                self.assertEqual(c.trace_from.outparams["ret1"], "val1")

                # test third connection which should be matching name only
                c = tp.connections[2]
                self.assertIsInstance(c, MatchingNameEdge)
                self.assertEqual(c.trace_from.fn_name, "bla.wut3")
                self.assertEqual(c.trace_to.fn_name, "bla.wut4")
                self.assertIn("output-value", c.trace_from.outparams)
                self.assertEqual(c.trace_from.outparams["output-value"], "val1")
                self.assertIn("output-value", c.trace_to.inparams)
                self.assertEqual(c.trace_to.inparams["output-value"], "val2")

                # test pruning
                tp.prune_connections()
                self.assertEqual(len(tp.connections), 3)

                # add another one that will be pruned
                t5 = Trace()
                t5.start()
                t5.set_input("bla.wut5", {"ret1": "val1"})
                t5.set_output("reqid5", "bla.wut5", {})
                t5.finish()
                tp.traces.append(t5)

                tp.find_connections()
                self.assertEqual(len(tp.connections), 5)
                c = tp.connections[-1]
                self.assertEqual(c.trace_from.fn_name, "bla.wut3")
                tp.prune_connections()
                self.assertEqual(len(tp.connections), 4)
                c = tp.connections[-1]
                self.assertEqual(c.trace_from.fn_name, "bla.wut")



import unittest


class TestTracer(unittest.TestCase):

    def setUp(self):
        try:
            from awstracer.tracer import Trace
        except Exception:
            self.fail("Trace import failed")
        try:
            t = Trace()
        except Exception:
            self.fail("Trace construction failed")
        self.t = t

    def test_trace_attrs(self):
        attrs = ("request_id", "fn_name", "inparams", "outparams", "ts_start", "ts_end")
        t = self.t
        for attr in attrs:
            self.assertIn(attr, dir(t))
        self.assertEqual(t.ts_start, None)
        self.assertEqual(t.ts_end, None)
        self.assertIsInstance(t.request_id, str)
        self.assertIsInstance(t.fn_name, str)
        self.assertIsInstance(t.inparams, dict)
        self.assertIsInstance(t.outparams, dict)

    def test_trace_start_finish(self):
        t = self.t
        for attr in ("start", "finish"):
            self.assertIn(attr, dir(t))
        t.start()
        import time
        time.sleep(.01)
        t.finish()
        diff = t.ts_end - t.ts_start
        self.assertEqual(diff.seconds, 0)
        self.assertGreaterEqual(diff.microseconds, 10)

    def test_trace_set_inputs(self):
        t = self.t
        for attr in ("set_input", "set_output"):
            self.assertIn(attr, dir(t))
        t.set_input("bla", {"wut": "bla"})
        self.assertEqual(t.fn_name, "bla")
        self.assertIn("wut", t.inparams)
        self.assertEqual(t.inparams["wut"], "bla")

        # test for different fn_name for output values
        with self.assertRaises(ValueError):
            t.set_output("blaaaaaal", "", {})
        t.set_output("reqid", "bla", {"wot": "bluh"})
        self.assertEqual(t.fn_name, "bla")
        self.assertEqual(t.request_id, "reqid")
        self.assertIn("wot", t.outparams)
        self.assertEqual(t.outparams["wot"], "bluh")

    def test_trace_tofrom_dicts(self):
        from awstracer.tracer import Trace
        t = self.t
        for attr in ("to_dict", "from_dict"):
            self.assertIn(attr, dir(t))
        d = t.to_dict()
        for n in ("request_id", "fn_name", "inparams", "outparams", "ts_start", "ts_end"):
            self.assertIn(n, d)
            self.assertEqual(d[n], getattr(t, n))
        t2 = Trace.from_dict(d)
        for n in ("request_id", "fn_name", "inparams", "outparams", "ts_start", "ts_end"):
            self.assertIn(n, dir(t2))
            self.assertEqual(getattr(t, n), getattr(t2, n))

    def test_trace_get_shell(self):
        from awstracer.tracer import Trace
        t = Trace()
        t.set_input("bla", {"wut": "wot"})
        for attr in ("get_shell_poc", "get_shell_var"):
            self.assertIn(attr, dir(t))
        with self.assertRaises(ValueError):
            # fn name is not correct
            t.get_shell_poc({})
        t.set_input("bla.blu", {"wut": "wot", "hi": "`\"escape'"})
        r = t.get_shell_poc()
        self.assertTrue(r.startswith("aws"))
        r = r.split(" ")
        self.assertEqual(len(r), 7)
        self.assertEqual(r[1], "bla")
        self.assertEqual(r[2], "blu")
        self.assertEqual(r[3], "--wut")
        self.assertEqual(r[4], "wot")
        self.assertEqual(r[5], "--hi")
        self.assertEqual(r[6], '\'`"escape\'"\'"\'\'')

        r = t.get_shell_var("BLA", "wut")
        self.assertEqual(len(r), 2)
        self.assertEqual(r[0], "--bl-a")
        self.assertEqual(r[1], "wut")

        r = t.get_shell_var("hi", "`\"arg{};'")
        self.assertEqual(len(r), 2)
        self.assertEqual(r[0], "--hi")
        self.assertEqual(r[1], "\'`\"arg{};\'\"\'\"\'\'")

    def test_event_capturer(self):
        try:
            from awstracer.tracer import EventCapturer
        except Exception:
            self.fail("error while importing eventcapturer")
        ev = EventCapturer()
        self.assertIn("trace", dir(ev))
        self.assertIn("events_captured", dir(ev))
        self.assertIn("emit", dir(ev))

        ev.emit("bla")
        self.assertIn("bla", ev.events_captured)
        with self.assertRaises(Exception):
            ev.emit("provide-client-params", **{})
            ev.emit("provide-client-params", **{"params": {}})
            ev.emit("provide-client-params.", **{"params": {}})
        ev.emit("provide-client-params.x", **{"params": {}})
        self.assertEqual(ev.trace.fn_name, "x")
        ev.emit("provide-client-params.xy", **{"params": {"test": "123"}})
        self.assertEqual(ev.trace.fn_name, "xy")
        self.assertIn("test", ev.trace.inparams)
        self.assertEqual(ev.trace.inparams["test"], "123")

        with self.assertRaises(Exception):
            ev.emit("after-call", **{})
            ev.emit("after-call.", **{})
            ev.emit("after-call.asdfx", **{"parsed": {}})
            ev.emit("after-call.asdfx", **{"parsed": {"ResponseMetadata": {}}})
            ev.emit("after-call.asdfx", **{"parsed": {"ResponseMetadata": {"RequestId": "req123"}}})

        # reset input parameters too to make them match
        ev.emit("provide-client-params.asdfx", **{"params": {}})
        ev.emit("after-call.asdfx", **{"parsed": {"ResponseMetadata": {"RequestId": "req123"}, "tro": "lll"}})
        self.assertEqual(ev.trace.fn_name, "asdfx")
        self.assertEqual(ev.trace.request_id, "req123")
        self.assertEqual(len(ev.trace.outparams), 1)
        self.assertIn("tro", ev.trace.outparams)
        self.assertEqual(ev.trace.outparams["tro"], "lll")

    def test_tracerunner(self):
        try:
            from awstracer.tracer import TraceRunner
            from awstracer.tracer import EventCapturer
        except Exception:
            self.fail("import failed for tracerunner")
        try:
            import awscli
        except Exception:
            self.fail("awscli import failed")

        self.assertIn("EnvironmentVariables", dir(awscli))
        self.assertIn("clidriver", dir(awscli))

        # establish that all the internals of AWS CLI haven't changed
        # such that the tracer _create_clidriver() most likely still
        # works
        try:
            import awscli.plugin
        except Exception:
            self.fail("awscli.plugin import failed")
        self.assertIn("load_plugins", dir(awscli.plugin))
        self.assertIn("CLIDriver", dir(awscli.clidriver))
        self.assertIn("_set_user_agent_for_session", dir(awscli.clidriver))

        try:
            import botocore
            import botocore.hooks
        except Exception:
            self.fail("botocore/botocore.hooks import failed")

        self.assertIn("session", dir(botocore))
        self.assertIn("Session", dir(botocore.session))

        s = botocore.session.Session(awscli.EnvironmentVariables)
        self.assertIn("register_component", dir(s))
        tr = TraceRunner()
        ev = EventCapturer()
        self.assertIn("run_aws_cmd", dir(tr))
        self.assertIn("_create_clidriver", dir(tr))

        # redirect stdout and stderr as the AWS CLI output is annoying and not necessary
        from contextlib import redirect_stderr, redirect_stdout
        import io
        with self.assertRaises(SystemExit):
            with redirect_stdout(io.StringIO()):
                with redirect_stderr(io.StringIO()):
                    tr.run_aws_cmd(["definitelynotexistingcommand"])
                    tr.run_aws_cmd(["s3", "definitelynotexistingcommand"])

        ret = tr._create_clidriver(ev)
        self.assertIsNotNone(ret)
        self.assertIsInstance(ret, awscli.clidriver.CLIDriver)



import io
import tempfile
import unittest
import uuid
from contextlib import redirect_stderr, redirect_stdout


class TestRecorder(unittest.TestCase):
    def test_options(self):
        try:
            from awstracer.recorder import opt_parser
        except Exception:
            self.fail("cannot import opt_parser")
        with self.assertRaises(SystemExit):
            with redirect_stdout(io.StringIO()):
                with redirect_stderr(io.StringIO()):
                    opt_parser([])
        ns = opt_parser(["--trace-file", "bla"])
        opts = ("prompt_color", "prompt_on_misc", "prompt_on_save", "enable_misc_cmd", "trace_file")
        for opt in opts:
            self.assertIn(opt, ns)
        self.assertEqual(ns.trace_file, "bla")
        # check default settings for options
        self.assertTrue(ns.prompt_color)
        self.assertTrue(ns.prompt_on_save)
        self.assertTrue(ns.prompt_on_misc)
        self.assertFalse(ns.enable_misc_cmd)
        ns = opt_parser(["--trace-file", "bla", "-c", "-n", "-s", "-d"])
        self.assertFalse(ns.prompt_color)
        self.assertFalse(ns.prompt_on_save)
        self.assertFalse(ns.prompt_on_misc)
        self.assertTrue(ns.enable_misc_cmd)

    def test_recorder_class(self):
        try:
            from awstracer.recorder import TraceRecorder
            from awstracer.tracer import TraceRunner
        except Exception:
            self.fail("cannot import TraceRecorder")
        with self.assertRaises(TypeError):
            tr = TraceRecorder()
        # give random filename
        tr = TraceRecorder(uuid.uuid4().hex)
        # prompt on save should be on by default
        self.assertIsInstance(tr, TraceRunner)
        self.assertTrue(tr.prompt_on_save)
        self.assertIn("run_aws_cmd", dir(tr))

    def test_recorder_file_arguments(self):
        try:
            from awstracer.recorder import TraceRecorder
        except Exception:
            self.fail("cannot import TraceRecorder")
        out = io.StringIO()
        with redirect_stdout(out):
            tr = TraceRecorder(uuid.uuid4().hex)
            tmp_fn = "/tmp/{}".format(uuid.uuid4().hex)
            tr.run_aws_cmd(["aws", "bla", "bla", "file://{}".format(tmp_fn)])
            # check for error message which should contain the file we can't read
            self.assertNotEqual(out.getvalue().find(tmp_fn), -1)

        # test actual reading from file
        tp = tempfile.NamedTemporaryFile()
        tp.write(b"hello w0rld")
        tp.seek(0)
        tr = TraceRecorder(uuid.uuid4().hex)
        args = tr.process_file_arguments(["aws", "bla", "bla", "file://{}".format(tp.name)])
        self.assertEqual(len(args), 4)
        self.assertEqual(args[3], "hello w0rld")



import setuptools
import re

name = "awstracer"

# We could use import obviously but we parse it as some python build systems
# otherwise pollute namespaces and we might end up with some annoying issues.
# See https://stackoverflow.com/a/7071358 for a discussion.
vfile = "src/{}/_version.py".format(name)
with open(vfile, "rt") as fd:
    verstrline = fd.read()
    regex = r"^__version__ = ['\"]([^'\"]*)['\"]"
    mo = re.search(regex, verstrline, re.M)
    if mo:
        version = mo.group(1)
    else:
        raise RuntimeError("Unable to find version string in %s." % (vfile,))

# load long description directly from the include markdown README
with open("README.md", "r") as fd:
    long_description = fd.read()

setuptools.setup(
    name=name,
    version=version,
    author="Anvil Secure",
    author_email="info@anvilsecure.com",
    description="AWS cli trace recorder and player",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/anvilsecure/awstracer",
    packages=setuptools.find_packages(where="src"),
    package_dir={"": "src"},
    keywords="AWS awscli cli tracer recorder player utility",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Operating System :: POSIX",
        "Operating System :: MacOS :: MacOS X",
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.6",
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: System Administrators",
    ],
    python_requires='>=3.6',
    entry_points={
        "console_scripts": ["awstrace-play=awstracer.player:main",
                            "awstrace-rec=awstracer.recorder:main"]
    },
    install_requires=[
        "awscli>=1.18.39",
        "botocore>=1.15.39"
    ],
    extras_require={
        "test": ["tox", "flake8"]
    }
)
